/*****************************************************************************

	Noise gate example and calibration application
	for the Electro-Smith Daisy

	Author: Laurent de Soras, 2020-2021


NoiseBleach is a 8-band noise gate. It is some kind of intermediate solution
between a standard noise gate and an FFT-based noise removal. It has more or
less the best of both worlds.

Each frequency band has its own threshold and gate time. The former can be
configured at any time, and the latter is automatic.

Even if the default configuration works, you should take the time to calibrate
the gate and find the optimal values for your hardware settings.

Compile with:
C_INCLUDES ?= -I.
C_DEFS ?= -DNDEBUG
OPT ?= -O3

You may have to add CPPFLAGS += --std=c++14 too, but this should work without
it on any recent compiler.

The code makes intensive use of debugging assert(), so remove them with
-DNDEBUG in production. A single channel uses about 14% of the CPU with a
block length of 48 samples. You can reduce the number of bands to save CPU,
but this will lower the noise efficiency.

This class was initially part of a larger code base, hence the weird
dependencies I tried to group and minimize. The full code can be found here:
https://github.com/EleonoreMizo/pedalevite/
It's not made for Daisy but the portable parts including the DSP bits work
fine, or require minute adjustment.


Calibration procedure
=====================

You'll need:

- The regular audio source for your setup (a guitar, a console output,
whatever). Of course you can do it without, it's just that it won't take the
source noise into account, only the Daisy environment.

- Something to record the output of the Daisy. No super spec is required,
just the bare minimum.

- An audio file editor to check the recording, preferably with a spectral
display (it's easier but not mandatory at all)

- And of course your Daisy hardware.

First, set the #define CALIBRATE_NOISE_BLEACH after the #include list in
main.cpp (this file). Compile and upload the program to your Daisy.

Run the test with an instrument plugged to the Daisy Seed left input. It
should remain as silent as possible (put a sock between the strings of your
guitar) but not muted at the volume pot.

Record the output and press the Reset button to start the run. Wait for
4 minutes and stop the recording.

Open the file in your audio editor. For simplicity, trim the file at the reset
"plop". This will be easier to measure the time offsets.

The 8 frequency bands are swept in ascending order, one after each other.
They starts from -144 dB and the threshold is increased by 6 dB every
2 seconds. There are 15 steps, so this is 30 s per band. Normally, each band
should start at full volume. At a given time, it starts to fade out step by
steps, then finally becomes silent.

For each band, find the moment where it becomes silent. It's easier to check
with a spectral display rather than a waveform. Subtract this moment from the
band time offset. Divide by 2 this duration in seconds to get a step index,
multiply by 6 to get a dB value, then subtract 144. This is your threshold
level in dB. Then convert it to a linear scale.

step      = (time_silence - time_band_beg) / 2
level_dB  = 6 * step - 144
level_lin = pow (10, level_db / 20)

Once you've done this for the 8 bands, report the value in the nzbl_thr_arr
array which is used to initialize the NoiseBleach objects. You can tweak the
value at your taste to let pass a bit of noise and increase slightly the
sustain.


--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://www.wtfpl.net/ for more details.

*Tab=3***********************************************************************/

#include "dzy/NoiseBleach.h"

#include "daisy_seed.h"
#include "hid/audio.h"

#include <algorithm>
#include <array>

#include <cmath>



// Replace #undef with #define to run the noise gate calibration procedure
#define CALIBRATE_NOISE_BLEACH



// Convenience function to convert dB to linear values
float db2lin (float db)
{
	return powf (10.f, db / 20.f);
}



// The number of input channels. Generally 1 or 2.
constexpr int max_nbr_chn = 2;

// Audio processing block size, in sample frames
constexpr int block_size  = 48;

// Please make sure to use one instance of NoiseBleach per channel.
std::array <dzy::NoiseBleach, max_nbr_chn> fx_nzbl;

// This is the calibration data.
const std::array <float, dzy::NoiseBleach::_nbr_bands> nzbl_thr_arr {{
	db2lin (10   * 6 - 144),
	db2lin ( 6   * 6 - 144),
	db2lin ( 7   * 6 - 144),
	db2lin ( 6.5 * 6 - 144),
	db2lin ( 7   * 6 - 144),
	db2lin ( 7   * 6 - 144),
	db2lin ( 8   * 6 - 144),
	db2lin ( 8.5 * 6 - 144)
}};

// We need a temporary buffer to store continuous channel data
float buffer [block_size];

daisy::DaisySeed hardware;
float sample_freq = 0; // Sampling rate, Hz



// The NoiseBleach interface is better suited for channel-packed data
// (stored with float **ptr). We use here interleaved channel data to show
// how it works with this format.
void AudioCallback (daisy::AudioHandle::InterleavingInputBuffer src_ptr, daisy::AudioHandle::InterleavingOutputBuffer dst_ptr, size_t bufsize)
{
	const int      nbr_spl = int (bufsize) >> 1;

#if ! defined (CALIBRATE_NOISE_BLEACH)

	// Normal run
	for (int chn = 0; chn < max_nbr_chn; ++chn)
	{
		// Input to buffer
		for (int i = 0; i < nbr_spl; ++i)
		{
			buffer [i] = src_ptr [i * 2 + chn];
		}

		// Filtering
		fx_nzbl [chn].process_block (buffer, buffer, nbr_spl);

		// Buffer to output
		for (int i = 0; i < nbr_spl; ++i)
		{
			dst_ptr [i * 2 + chn] = buffer [i];
		}
	}

#else // CALIBRATE_NOISE_BLEACH

	// Calibration program
	static int     time_spl = 0;
	static int     cur_sec  = 0;
	if (time_spl >= cur_sec * sample_freq)
	{
		constexpr int  db_min    = -144; // dB
		constexpr int  db_max    =  -60; // dB
		constexpr int  db_step   =    6; // dB
		const int      nbr_steps = 1 + (db_max - db_min + db_step - 1) / db_step;
		const int      idx_glob = cur_sec / 2;
		const int      step_idx = idx_glob % nbr_steps;
		const int      band_idx = idx_glob / nbr_steps;
		if (band_idx < dzy::NoiseBleach::_nbr_bands)
		{
			const float    thr = db2lin (step_idx * 6.0 + db_min);
			fx_nzbl [0].set_threshold (band_idx, thr);
		}

		cur_sec += 2;
	}

	for (int i = 0; i < nbr_spl; ++i)
	{
		buffer [i] = src_ptr [i * 2];
	}

	fx_nzbl [0].process_block (buffer, buffer, nbr_spl);

	// Amplifies everything to hell and hardclips to -15 dBFS to avoid damaging
	// ears or monitors if some "normal" sound is present at the input.
	constexpr float   gain     = 1000.f;
	constexpr float   clip_val = 0.18f;
	for (int i = 0; i < nbr_spl; ++i)
	{
		float          x = buffer [i];
		x *= gain;
		x  = std::min (std::max (x, -clip_val), +clip_val);
		dst_ptr [i * 2    ] = x;
		dst_ptr [i * 2 + 1] = x;
	}

	time_spl += nbr_spl;

#endif // CALIBRATE_NOISE_BLEACH

}



int main (int argc, char *argv [])
{
	hardware.Configure ();
	hardware.Init ();

	sample_freq = hardware.AudioSampleRate ();
	hardware.SetAudioBlockSize (block_size);

	// Noise gate setup.
	for (int chn = 0; chn < max_nbr_chn; ++chn)
	{
		fx_nzbl [chn].init (sample_freq);
		for (int b = 0; b < dzy::NoiseBleach::_nbr_bands; ++b)
		{
#if defined (CALIBRATE_NOISE_BLEACH)
			// A really high threshold, so we make the input totally silent
			// when there's only the background noise.
			const float    thr = 1e-3f;
#else
			const float    thr = nzbl_thr_arr [b];
#endif
			fx_nzbl [chn].set_threshold (b, thr);
		}
	}

	hardware.StartAudio (AudioCallback);

	// Loop forever
	for (;;)
	{
		daisy::System::Delay (10);
	}

	return 0;
}
