/*
Author: Laurent de Soras

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://www.wtfpl.net/ for more details.
*/



#include "noisebleach_dependencies.h"

#include <algorithm>

#include <cassert>
#include <cmath>



namespace mfx
{
namespace dsp
{
namespace dyn
{



class EnvHelper
{
public:

	static inline double
	               compute_env_coef_simple (double t, double fs) noexcept;

}; // class EnvHelper

// Approximates c = 1 - exp (-1 / (fs * t))
// Coefficient in ]0 ; 1]
double	EnvHelper::compute_env_coef_simple (double t, double fs) noexcept
{
	assert (fs > 0);
	assert (t >= 0);

	const double   p = t * fs;
	const double   r = 1.90625;
	const double   u = r / (1 + (3 * r) * p);
	const double   v = std::max (1 - u, 0.0);
	const double   c = 1 - v * v * v;

	return c;
}



void	EnvFollowerRms::set_sample_freq (double fs) noexcept
{
	assert (fs > 0);

	_sample_freq = float (fs);
	update_parameters ();
}

void	EnvFollowerRms::set_times (float at, float rt) noexcept
{
	assert (at >= 0);
	assert (rt >= 0);

	_time_a = at;
	_time_r = rt;
	update_parameters ();
}

// Input is not squared
// Output is not square-rooted
float	EnvFollowerRms::analyse_block_raw_cst (float x2, int nbr_spl) noexcept
{
	assert (x2 >= 0);
	assert (nbr_spl > 0);

	const float    delta = x2 - _state;
	const float    coef  = (delta >= 0) ? _coef_a : _coef_r;

	const float    coef_block =
		  (nbr_spl < 100000)
		? 1.f - fstb::ipowp (1.f - coef, nbr_spl)
		: 1.f - powf (1.f - coef, nbr_spl);

	_state += delta * coef_block;

	return _state;
}

void	EnvFollowerRms::clear_buffers () noexcept
{
	_state = 0;
}

void	EnvFollowerRms::update_parameters () noexcept
{
	_coef_a = float (
		EnvHelper::compute_env_coef_simple (_time_a, _sample_freq)
	);
	_coef_r = float (
		EnvHelper::compute_env_coef_simple (_time_r, _sample_freq)
	);
}



}  // namespace dyn



namespace iir
{

/*
==============================================================================
Name: process_block
Description:
	Filters a block of samples. Can work in-place.
Input parameters:
	- in_ptr: Input sample block
	- nbr_spl: Number of samples to process. >= 0
Output parameters:
	- out_ptr: Output sample block
Throws: Nothing
==============================================================================
*/

void	Biquad::process_block (float dst_ptr [], const float src_ptr [], int nbr_spl) noexcept
{
	assert (dst_ptr != nullptr);
	assert (src_ptr != nullptr);
	assert (nbr_spl > 0);

	// If we're not on an even boudary, we process a single sample.
	if (_mem_pos != 0)
	{
		*dst_ptr = process_sample (*src_ptr);
		++ src_ptr;
		++ dst_ptr;
		-- nbr_spl;

		if (nbr_spl <= 0)
		{
			return;
		}
	}

	int            half_nbr_spl = nbr_spl >> 1;
	int            index = 0;
	if (half_nbr_spl > 0)
	{
		float          mem_x [2] { _mem_x [0], _mem_x [1] };
		float          mem_y [2] { _mem_y [0], _mem_y [1] };

		do
		{
			// First sample
			float          x = src_ptr [index];
			mem_y [1] =      _eq_z_b [0] *     x
			            + (  _eq_z_b [1] * mem_x [0]
			               + _eq_z_b [2] * mem_x [1])
			            - (  _eq_z_a [1] * mem_y [0]
			               + _eq_z_a [2] * mem_y [1]);

			mem_x [1] = x;
			dst_ptr [index    ] = mem_y [1];

			// Second sample
			x = src_ptr [index + 1];
			mem_y [0] =      _eq_z_b [0] *     x
			            + (  _eq_z_b [1] * mem_x [1]
			               + _eq_z_b [2] * mem_x [0])
			            - (  _eq_z_a [1] * mem_y [1]
			               + _eq_z_a [2] * mem_y [0]);

			mem_x [0] = x;
			dst_ptr [index + 1] = mem_y [0];

			index += 2;
			-- half_nbr_spl;
		}
		while (half_nbr_spl > 0);

		_mem_x [0] = mem_x [0];
		_mem_x [1] = mem_x [1];
		_mem_y [0] = mem_y [0];
		_mem_y [1] = mem_y [1];
	}

	// If number of samples was odd, there is one more to process.
	if ((nbr_spl & 1) > 0)
	{
		dst_ptr [index] = process_sample (src_ptr [index]);
	}
}

void	Biquad::clear_buffers ()
{
	_mem_x [0] = 0;
	_mem_x [1] = 0;
	_mem_y [0] = 0;
	_mem_y [1] = 0;
	_mem_pos   = 0;
}



void	TransSZBilin::map_s_to_z_approx (float z_eq_b [3], float z_eq_a [3], const float s_eq_b [3], const float s_eq_a [3], float k) noexcept
{
	assert (z_eq_b != nullptr);
	assert (z_eq_a != nullptr);
	assert (s_eq_b != nullptr);
	assert (s_eq_a != nullptr);

	const float    kk = k*k;

	const float    b1k  = s_eq_b [1] * k;
	const float    a1k  = s_eq_a [1] * k;
	const float    b2kk = s_eq_b [2] * kk;
	const float    a2kk = s_eq_a [2] * kk;
	const float    b2kk_plus_b0 = b2kk + s_eq_b [0];
	const float    a2kk_plus_a0 = a2kk + s_eq_a [0];
	const float    b0z = b2kk_plus_b0 + b1k;
	const float    a0z = a2kk_plus_a0 + a1k;
	const float    b2z = b2kk_plus_b0 - b1k;
	const float    a2z = a2kk_plus_a0 - a1k;
	const float    b1z = 2 * (s_eq_b [0] - b2kk);
	const float    a1z = 2 * (s_eq_a [0] - a2kk);

	// IIR coefficients
	assert (a0z != 0);
	const float    mult = 1 / a0z;

	z_eq_b [0] = b0z * mult;
	z_eq_b [1] = b1z * mult;
	z_eq_b [2] = b2z * mult;

	z_eq_a [0] = 1;
	z_eq_a [1] = a1z * mult;
	z_eq_a [2] = a2z * mult;
}



}  // namespace iir



namespace mix
{



void	Fpu::scale_1_vlr (float data_ptr [], int nbr_spl, float s_vol, float e_vol) noexcept
{
	assert (data_ptr != nullptr);
	assert (nbr_spl > 0);

	const float		step = (e_vol - s_vol) * fstb::rcp_uint <float> (nbr_spl);
	int				pos = 0;
	do
	{
		data_ptr [pos] *= s_vol;
		s_vol += step;
		++ pos;
	}
	while (pos < nbr_spl);
}

void	Fpu::copy_2_1 (float out_ptr [], const float in_1_ptr [], const float in_2_ptr [], int nbr_spl) noexcept
{
	assert (out_ptr != nullptr);
	assert (in_1_ptr != nullptr);
	assert (in_2_ptr != nullptr);
	assert (nbr_spl > 0);

	int				pos = 0;
	do
	{
		out_ptr [pos] = in_1_ptr [pos] + in_2_ptr [pos];
		++ pos;
	}
	while (pos < nbr_spl);
}

void	Fpu::mix_1_1 (float out_ptr [], const float in_ptr [], int nbr_spl) noexcept
{
	assert (out_ptr != nullptr);
	assert (in_ptr != nullptr);
	assert (nbr_spl > 0);

	int				pos = 0;
	do
	{
		out_ptr [pos] += in_ptr [pos];
		++ pos;
	}
	while (pos < nbr_spl);
}



}  // namespace mix
}  // namespace dsp
}  // namespace mfx
